#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: ngamsRestoreCmd.py,v 1.3 2007/05/31 15:54:38 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  04/05/2007  Created
#

"""
Contains code for handling the RESTORE Command.
"""

import pydoc
import pcc, PccUtTime
from ngams import *
import ngamsLib, ngamsHighLevelLib, ngamsFileInfo, ngamsNotification
import ngamsFileUtils, ngamsDiskUtils, ngamsArchiveUtils
import ngamsStatus, ngamsDapiStatus


def restoreFile(srvObj,
                reqPropsObj,
                httpRef,
                fileInfoObj):
    """
    Handle RESTORE Command.

    srvObj:         Reference to NG/AMS server class object (ngamsServer).
    
    reqPropsObj:    Request Property object to keep track of actions done
                    during the request handling (ngamsReqProps).
        
    httpRef:        Reference to the HTTP request handler
                    object (ngamsHttpRequestHandler).

    fileInfoObj:    File info object with information about the file
                    to restore (ngamsFileInfo).
        
    Returns:        Void.
    """
    info(4,"Entering restoreFile() ...")

    trgDiskInfo = None
    diskExemptList = [fileInfoObj.getDiskId()]
    while (True):
        try:
            tmpTrgDiskInfo = ngamsDiskUtils.\
                             findTargetDisk(srvObj.getDb(),srvObj.getCfg(),
                                            fileInfoObj.getFormat(), 0,
                                            diskExemptList)
            if (srvObj.getDb().fileInDb(tmpTrgDiskInfo.getDiskId(),
                                        fileInfoObj.getFileId(),
                                        fileInfoObj.getFileVersion())):
                diskExemptList.append(tmpTrgDiskInfo.getDiskId())
            else:
                trgDiskInfo = tmpTrgDiskInfo
                break
        except Exception, e:
            msg = "Error locating target disk for RESTORE Command. Error: %s"
            msg = msg % str(e)
            raise Exception, msg

    # Generate staging filename, receive data.
    reqPropsObj.setTargDiskInfo(trgDiskInfo)
    storageSetId = trgDiskInfo.getStorageSetId()
    tmpStagingFilename, stagingFilename,\
                        tmpReqPropsFilename,\
                        reqPropsFilename = ngamsHighLevelLib.\
                        genStagingFilename(srvObj.getCfg(), reqPropsObj,
                                           srvObj.getDiskDic(),
                                           storageSetId,
                                           reqPropsObj.getFileUri(),
                                           genTmpFiles=1)
    try:
        # Save the data into the Temp. Staging File.
        ioTime = ngamsHighLevelLib.saveInStagingFile(srvObj.getCfg(),
                                                     reqPropsObj,
                                                     stagingFilename,
                                                     trgDiskInfo)
        # Synchronize the file caches to ensure the files have been stored
        # on the disk and check that the files are accessible.
        timer = PccUtTime.Timer()
        ngamsFileUtils.syncCachesCheckFiles(srvObj, [stagingFilename])
        ioTime += timer.stop()

        # If a checksum value is available in NGAS Files, check the checksum
        # of the file.
        if (fileInfoObj.getChecksumPlugIn() and fileInfoObj.getChecksum()):
            dcpi = fileInfoObj.getChecksumPlugIn()
            exec "import " + dcpi
            checksumStgFile = eval(dcpi + "." + dcpi +\
                                   "(srvObj, stagingFilename)")
            if (checksumStgFile != fileInfoObj.getChecksum()):
                msg = "Illegal checksum for file to restore. Reference " +\
                      "file: %s/%s/%s" % (fileInfoObj.getDiskId(),
                                          fileInfoObj.getFileId(),
                                          str(fileInfoObj.getFileVersion()))
                warning(msg)
                raise Exception, msg
        else:
            msg = "No checksum or checksum plug-in specified for source " +\
                  "file for RESTORE Command: %s/%s/%s - skipping checksum " +\
                  "check"
            info(1,msg % (fileInfoObj.getDiskId(),
                          fileInfoObj.getFileId(),
                          str(fileInfoObj.getFileVersion())))
        
        # Generate the DB information for the restored file (=target file).
        newFileInfoObj = fileInfoObj.clone().\
                         setDiskId(trgDiskInfo.getDiskId()).\
                         setCreationDateFromSecs(int(time.time() + 0.5))
                       
        # Create the target file.
        targetFilename = os.path.normpath("%s/%s" %\
                                          (trgDiskInfo.getMountPoint(),
                                           newFileInfoObj.getFilename()))
        info(3,"Move Restore Staging File to final destination: %s->%s ..." %\
             (stagingFilename, targetFilename))
        timer.start()
        mvFile(stagingFilename, targetFilename)
        ioTime += timer.stop()
        info(3,"Moved Restore Staging File to final destination: %s->%s" %\
             (stagingFilename, targetFilename))

        # Update the DB with the information about the new file.
        # Update information for Main File/Disk in DB.
        newFileInfoObj.write(srvObj.getDb())
        diskSpace = getDiskSpaceAvail(trgDiskInfo.getMountPoint())
        newSize = trgDiskInfo.getBytesStored() + newFileInfoObj.getFileSize()
        ioTime = trgDiskInfo.getTotalDiskWriteTime() + ioTime
        trgDiskInfo.\
                      setNumberOfFiles(trgDiskInfo.getNumberOfFiles() + 1).\
                      setAvailableMb(diskSpace).\
                      setBytesStored(newSize).\
                      setTotalDiskWriteTime(ioTime).\
                      write(srvObj.getDb())

        # Ensure that the restored file is readonly.
        ngamsLib.makeFileReadOnly(targetFilename)
        
        # Add the file info object for the target file in the disk info object.
        trgDiskInfo.addFileObj(newFileInfoObj)

        reqPropsObj.incIoTime(ioTime)
    except Exception, e:
        # Remove the staging file(s).
        rmFile(stagingFilename)
        raise e
    
    info(4,"Leaving restoreFile()")
    return trgDiskInfo
    

def handleCmdRestore(srvObj,
                     reqPropsObj,
                     httpRef):
    """
    Handle RESTORE Command.
        
    srvObj:         Reference to NG/AMS server class object (ngamsServer).
    
    reqPropsObj:    Request Property object to keep track of actions done
                    during the request handling (ngamsReqProps).
        
    httpRef:        Reference to the HTTP request handler
                    object (ngamsHttpRequestHandler).
        
    Returns:        Void.
    """
    info(4,"Entering handleCmdRestore() ...")

    # Is this NG/AMS permitted to handle Restore (=~Archive) Requests?
    if (not srvObj.getCfg().getAllowArchiveReq()):
        errMsg = genLog("NGAMS_ER_ILL_REQ", ["Archive"])
        raise Exception, errMsg
    srvObj.checkSetState("Archive Request", [NGAMS_ONLINE_STATE],
                         [NGAMS_IDLE_SUBSTATE, NGAMS_BUSY_SUBSTATE],
                         NGAMS_ONLINE_STATE, NGAMS_BUSY_SUBSTATE)

    # Get command line parameters.
    fileId      = None
    fileVersion = None
    if (reqPropsObj.hasHttpPar("file_id")):
        fileId = reqPropsObj.getHttpPar("file_id")
    if (reqPropsObj.hasHttpPar("file_version")):
        fileVersion = reqPropsObj.getHttpPar("file_version")
    if ((not fileId) or (not fileVersion) or
        (not reqPropsObj.getFileUri())):
        errMsg = genLog("NGAMS_ER_MIS_PAR",
                        ["file_id=(ID)&file_version=(Version)&" +\
                         "filename=(Name)", NGAMS_RESTORE_CMD])
        raise Exception, errMsg

    # Ensure that the file info for the given File ID/Version is found in
    # the DB.
    fileInfoList = srvObj.getDb().getFileInfoFromFileId(fileId, fileVersion,
                                                        dbCursor=0, order=0)
    if (fileInfoList == []):
        raise Exception, "No info found for File ID/Version: %s/%s" %\
              (fileId, fileVersion)
    fileInfoObj = ngamsFileInfo.ngamsFileInfo()
    fileInfoObj.unpackSqlResult(fileInfoList[0])

    # Check if should act as proxy for the Restore Request.
    if (ngamsArchiveUtils.negotiateTarget(srvObj, reqPropsObj, httpRef,
                                          fileInfoObj.getFormat())):
        return

    # Now, go ahead, restore the file.
    try:
        diskinfoObj = restoreFile(srvObj, reqPropsObj, httpRef, fileInfoObj)
        msg = "Successfully handled Restore Request for " +\
              "data object with File ID/Version: %s/%s" %\
              (fileId, str(fileVersion))
        info(1,msg)
        srvObj.setSubState(NGAMS_IDLE_SUBSTATE)
        srvObj.ingestReply(reqPropsObj, httpRef, NGAMS_HTTP_SUCCESS,
                           NGAMS_SUCCESS, msg, diskinfoObj)
    except Exception, e:
        errMsg = genLog("NGAMS_ER_RESTORE_REQ",
                        [fileId, int(fileVersion), str(e)])
        error(errMsg)
        ngamsNotification.notify(srvObj.getCfg(), NGAMS_NOTIF_ERROR,
                                 "PROBLEM RESTORE COMMAND HANDLING", errMsg)
        raise Exception, errMsg

    info(4,"Leaving handleCmdRestore()")


# EOF
